//==============================================================================
// FastBitmapCore.pas
//------------------------------------------------------------------------------
// Historique :
// 18/11/16 - BeanzMaster - Creation
//------------------------------------------------------------------------------
// Description :
// Composant graphique optimisé pour l'affichage d'un Bitmap en 32bits
// Creation suite à une discussion sur :
// http://www.developpez.net/forums/d1610633/autres-langages/pascal/lazarus/info-font-graphisme-pointu/
//------------------------------------------------------------------------------
// Credits :
// Composant original :  FCL Compact Image Unit (c) 2016, Paul F. Michell.
//------------------------------------------------------------------------------
//==============================================================================
Unit FastBitmapCore;
//==============================================================================

{$I beanz.inc}

//==============================================================================
Interface

Uses
  Classes, fpImage,Graphics;

//==============================================================================
Type
  TColorFormat = (cfRGBA,cfBGRA);

//------------------------------------------------------------------------------
// Les macros, ici permettent de choisir la spécification des données du bitmap
// BGRA ou RGBA suivant votre OS, ou comment vous désirez utiliser le bitmap
// Pour Windows le format d'affichage par defaut est BGRA
// cf fichier beanz.inc pour modifier les options de compilations
//------------------------------------------------------------------------------
Type
  TCompactColor = Packed Record
    {$IFDEF BGRAIMAGE}
    Blue, Green, Red, Alpha: Byte;
    {$ELSE}
    Red, Green, Blue, Alpha: Byte;
    {$ENDIF}
  End;
  TCompactColorPointer = ^TCompactColor;

//------------------------------------------------------------------------------
const
  FastPixelSize = SizeOf(TCompactColor);

//===[ TFASTBITMAP ]============================================================
// Object optimizé pour utilisé un Bitmap en 32bit (RGBA)
// Pour l'affichage et le chargement de fichiers bitmap on utilise un objet
// descendant de TCustomFastBitmapHandler (TLCLFastBitmapHandler) ou le composant
// visuel TFastImage
// On peux également, directement copier les données accessibles via la propriété
// "Data" avec la procedure Move(Src,Dest,Size) ou parcourir les données via
// Pixel[x,y],Colors[x,y] ou accéder aux données à l'aide de pointeurs
//------------------------------------------------------------------------------
Type
  TFastBitmap = class(TFPCustomImage)
  private
    FBytesPerPixel : Integer;
    FBytesPerLine : Integer;
    FColorFormat : TColorFormat;

    FFastScanLine : array of integer;
    FOnChange: TNotifyEvent;

    procedure setColorFormat(value:TColorFormat);

    procedure SetRawPixel (x,y:integer; Color: TCompactColor); //inline;
    function GetRawPixel (x,y:integer) : TCompactColor;
    procedure ComputeFastScanLine;
  Protected
    FPixelsData: TCompactColorPointer;

    procedure DoOnChange(Sender: TObject); virtual;
    Function GetInternalColor(x, y: Integer): TFPColor; Override;
    Function GetInternalPixel({%H-}x, {%H-}y: Integer): Integer; Override;
    Procedure SetInternalColor (x, y: Integer; Const Value: TFPColor); Override;
    Procedure SetInternalPixel({%H-}x, {%H-}y: Integer; {%H-}Value: Integer); Override;
    Function GetFastScanLineAddr(y:integer):Integer;
  Public
    constructor Create(AWidth, AHeight: integer); override;
    Destructor Destroy; Override;
    Procedure Clear(Color: TCompactColor);
    Procedure SetSize(AWidth, AHeight: Integer); Override;
    function GetPixelAddress(X, Y: Integer): TCompactColorPointer; inline;
    procedure Changed();
    Property Data: TCompactColorPointer Read FPixelsData Write FPixelsData;

    property BytesPerLine : Integer read FBytesPerLine write FBytesPerLine;
    property BytePerPixel : Integer read FBytesPerPixel;
    property Pixel[x,y:integer] : TCompactColor read GetRawPixel write SetRawPixel;
  published
    property ColorFormat : TColorFormat Read FColorFormat write setColorFormat;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;

    property Width;
    property Height;
    property OnProgress;
  End;

//==============================================================================

Function RGBToIntColor(r,g,b:byte):Integer;
Function CompactColor(r,g,b,a:byte):TCompactColor;

//==============================================================================

Implementation

//==============================================================================
uses
  sysUtils{$IFDEF DEBUG_ON}, Dialogs {$ENDIF};
//==============================================================================

//------------------------------------------------------------------------------
// Renvoie une couleur au format TCompacColor
//------------------------------------------------------------------------------
Function CompactColor(r,g,b,a:byte):TCompactColor;
begin
  with result do
  begin
    red := r;
    blue := g;
    green := b;
    alpha := a;
  end;
end;

//------------------------------------------------------------------------------
// Convertion couleur RGB 24bit vers un "Integer"
//------------------------------------------------------------------------------
Function RGBToIntColor(r,g,b:byte):Integer;
begin
  result:=(r or (g shl 8) or (b shl 16)) and $ffffff;
end;

//------------------------------------------------------------------------------
// Inverse les composantes "Blue" et "Red" d'un TCompactColor
//------------------------------------------------------------------------------
function SwapBRComponent(Value: Cardinal): Cardinal;
begin
//  Result := (Value and $00ff00) or ((Value shr 16) and $ff) or ((Value and $ff) shl 16);
  Result := Value;
  TCompactColor(Result).Red := TCompactColor(Value).Blue;
  TCompactColor(Result).Blue := TCompactColor(Value).Red;
end;


//===[ TFASTBITMAP ]============================================================

//------------------------------------------------------------------------------
// Creation de TFastBitmap en spécifiant sa largeur et hauteur
//------------------------------------------------------------------------------
Constructor TFastBitmap.Create(AWidth, AHeight: integer);
Begin
  inherited Create(AWidth, AHeight);
  ReAllocMem(FPixelsData, FastPixelSize*AWidth*AHeight);
  FBytesPerPixel := FastPixelSize;
  FBytesPerLine := FastPixelSize * AWidth;
  UsePalette:=false;
end;

//------------------------------------------------------------------------------
// Destruction de TFastBitmap
//------------------------------------------------------------------------------
Destructor TFastBitmap.Destroy;
Begin
  ReAllocMem(FPixelsData, 0);
  FreeMem(FPixelsData);
  SetLength(FFastScanLine,0);
  FOnChange:=nil;
  Inherited Destroy;
End;

//------------------------------------------------------------------------------
// Change la taille de TFastBitmap
//------------------------------------------------------------------------------
Procedure TFastBitmap.SetSize(AWidth, AHeight: Integer);
Begin
  If (AWidth<>Width) Or (AHeight<>Height) Then
    Begin
      ReAllocMem(FPixelsData, FastPixelSize*AWidth*AHeight);
      FBytesPerLine := FastPixelSize * AWidth;
      Inherited SetSize(AWidth, AHeight);
      {$IFDEF DEBUG_ON}
      ShowMessage('FastBitmapCore.SetSize : '+inttostr(Width)+'x'+inttostr(Height));
      {$ENDIF}
      SetLength(FFastScanLine,Height);
      ComputeFastScanLine;
      DoOnChange(self);
    End;
End;

//------------------------------------------------------------------------------
// Definit le format de couleur cfRGBA ou cfBGRA (default)
//------------------------------------------------------------------------------
Procedure TFastBitmap.setColorFormat(value:TColorFormat);
Begin
  if (FColorFormat = Value) then exit;
  FColorFormat := Value;
  DoOnChange(self);
End;

//------------------------------------------------------------------------------
// Fonction interne à TFastBitmap, calcul des "scanlines"
// On se sert du tableau FFascanLine pour optimiser l'acces à un pixel.
// On evite ainsi une multiplication lorsque l'on doit modifier plusieurs pixels en une fois
//------------------------------------------------------------------------------
procedure TFastBitmap.ComputeFastScanLine;
Var
  y:Integer;
begin
  if BytesPerLine=0 then Exit;
  For y:=0 to Height-1 do
  begin
    FFastScanLine[y]:=y*Width; //FBytesPerLine
  end;
end;

//------------------------------------------------------------------------------
// Retourne le pointeur de type TCompactColorPointer à la position x,y
//------------------------------------------------------------------------------
function TFastBitmap.GetPixelAddress(X, Y: Integer): TCompactColorPointer;
begin
  Result := TCompactColorPointer(FPixelsData + (Y * Width + X));
end;

//------------------------------------------------------------------------------
// Retourne l'adresse physique d'une ligne dans notre bitmap
//------------------------------------------------------------------------------
Function TFastBitmap.GetFastScanLineAddr(y:integer):Integer;
begin
  result:=FFastScanLine[y];
end;

//------------------------------------------------------------------------------
// Non utilisé ici car on fonctionne en 32bit
// Cette procedure est utilisé pour des images avec une palette
// Retourne l'index de la couleur d'un pixel dans la palette
// cf : TFPCustomImage pour plus d'informations
//------------------------------------------------------------------------------
Function TFastBitmap.GetInternalPixel(x, y: Integer): Integer;
begin
  {$MESSAGE warn : GetInternalPixel Not used here }
  result:=0;

end;
//------------------------------------------------------------------------------
// Non utilisé ici car on fonctionne en 32bit
// Cette procedure est utilisé pour des images avec une palette
// Definit la couleur d'un pixel avec la couleur d'index Value
// presente dans la palette si définie
// cf : TFPCustomImage pour plus d'informations
//------------------------------------------------------------------------------
Procedure TFastBitmap.SetInternalPixel(x, y: Integer; Value: Integer);
begin
  {$MESSAGE warn : SetInternalPixel Not used here }
end;


//------------------------------------------------------------------------------
// Retourne une couleur de type TCompactColor à la position x,y
//------------------------------------------------------------------------------
Function TFastBitmap.GetRawPixel(x, y: Integer): TCompactColor;
Var
  xx,yy:integer;
begin
  yy:=FFastScanLine[y];
  xx:=yy+x;
  //Result:=FPixelsData[xx];
  Result:=TCompactColorPointer(FPixelsData+xx)^;
End;

//------------------------------------------------------------------------------
// Definie la couleur d'un pixel à la position x,y avec une couleur de type TCompactColor
//------------------------------------------------------------------------------
Procedure TFastBitmap.SetRawPixel(x, y: Integer; Color: TCompactColor);
Var
  xx,yy:integer;
begin
  //yy:=GetFastScanLineAddr(y); // Moins rapide que la ligne ci-dessous car la "porté" est plus grande
  yy:=FFastScanLine[y]; // yy:=y*width; petite optimisation on evite ainsi une multiplication
  xx:=yy+x;
  // Les 2 fonctions ci-dessous ce valent, la 1ere etant une peu plus performante d'un poil
  TCompactColorPointer(FPixelsData+xx)^:=Color;
  //FPixelsData[xx] := Color;
End;

//------------------------------------------------------------------------------
// Retourne une couleur de type FPColor à la position x,y
//------------------------------------------------------------------------------
Function TFastBitmap.GetInternalColor(x, y: Integer): TFPColor;
Var
  Color: TCompactColor;
Begin
  Color := FPixelsData[x+y*Width];
  With Color Do
    Begin
      Result.Red := (Red ShL 8)+Red;
      Result.Green := (Green ShL 8)+Green;
      Result.Blue := (Blue ShL 8)+Blue;
      Result.Alpha := (Alpha ShL 8)+Alpha;
    End;
End;

//------------------------------------------------------------------------------
// Definie la couleur d'un pixel à la position x,y avec une couleur de type TFPColor
//------------------------------------------------------------------------------
Procedure TFastBitmap.SetInternalColor(x, y: Integer; Const Value: TFPColor);
Var
  C:TCompactColor;
begin
  C:= CompactColor(byte(Value.red),byte(Value.green),byte(Value.blue),byte(Value.alpha));
  setRawPixel(x,y,C);
End;

//------------------------------------------------------------------------------
// Efface le bitmap avec une couleur de type TFPColor
//------------------------------------------------------------------------------
Procedure TFastBitmap.Clear(Color: TCompactColor);
Begin
  FillDWord(FPixelsData^, Width*Height, DWord(Color));
End;

//------------------------------------------------------------------------------
// Procedure Interne de notification en cas de modification du bitmap
//------------------------------------------------------------------------------
procedure TFastBitmap.DoOnChange(Sender:TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

//------------------------------------------------------------------------------
// Notification que le bitmap à changé
//------------------------------------------------------------------------------
procedure TFastBitmap.Changed();
begin
  DoOnChange(self);
end;


//==============================================================================

End.

